program Mesh2D
  use sep
  use util_
  use meshing2D
  implicit none
  
  type(myaxis) :: ax,ay,af
  integer              :: nt,ireg
  real                 :: dt,eps,gscale
  real   , allocatable :: f(:,:,:)
  complex, allocatable :: s(:,:)
  character(len=128)   :: name
  logical              :: verbose
  call sep_init()
  call hinaxis(ax,1)
  call hinaxis(ay,2)
  name="monitor"
  call ainaxis(af,3,name)
  call from_param("nt",nt,100) 
  call from_param("dt",dt,1.)
  call from_param("eps",eps,1.)
  call from_param("verbose",verbose,.false.)
  call from_param("gscale",gscale, 1.)
  call from_param("ireg",ireg,1)
  call houaxis(ay,1)
  call houaxis(ax,2) 
  
  !! Output the determinant
  call auxputch("n1","i",ay%n,"monitor.H")
  call auxputch("n2","i",ax%n,"monitor.H")
  call auxputch("n3","i",af%n,"monitor.H")
  call auxputch("o1","r",ay%o,"monitor.H")
  call auxputch("o2","r",ax%o,"monitor.H")
  call auxputch("o3","r",af%o,"monitor.H")
  call auxputch("d1","r",ay%d,"monitor.H")
  call auxputch("d2","r",ax%d,"monitor.H")
  call auxputch("d3","r",af%d,"monitor.H")

  !! Output coeff a1
  call auxputch("n1","i",ay%n,"a1.H")
  call auxputch("n2","i",ax%n,"a1.H")
  call auxputch("o1","r",1.,"a1.H")
  call auxputch("o2","r",1.,"a1.H")
  call auxputch("d1","r",1.,"a1.H")
  call auxputch("d2","r",1.,"a1.H")
  !! Output coeff a4
  call auxputch("n1","i",ay%n,"a4.H")
  call auxputch("n2","i",ax%n,"a4.H")
  call auxputch("o1","r",1.,"a4.H")
  call auxputch("o2","r",1.,"a4.H")
  call auxputch("d1","r",1.,"a4.H")
  call auxputch("d2","r",1.,"a4.H")
  !! Output coeff a5
  call auxputch("n1","i",ay%n,"a5.H")
  call auxputch("n2","i",ax%n,"a5.H")
  call auxputch("o1","r",1.,"a5.H")
  call auxputch("o2","r",1.,"a5.H")
  call auxputch("d1","r",1.,"a5.H")
  call auxputch("d2","r",1.,"a5.H")
  call sep_close()
  
  if (verbose) then
     write(0,*) "File         ","Number    ","Offset       ","Delta"  
     write(0,*) "x     ", ax%n,ax%o,ax%d
     write(0,*) "y     ", ay%n,ay%o,ay%d
     write(0,*) "monitor", af%n,af%o,af%d
  end if

  !! Allocate Memory
  allocate ( f(ax%n,ay%n,af%n),s(ax%n,ay%n) )
  
  !! Initialize Module
  call meshing2D_init(ax,ay,af,eps,nt,dt,verbose,gscale,ireg)

  !! Blended function or old coordinates
  call sep_read(s          )
  call sep_read(f,"monitor") 
  
  call runmeshing2D(s,f)

  call sep_write(s)
  
  deallocate (f,s)
  call exit()
end program Mesh2D
