module meshing2D
  use sep
  use util_
  use igrad1
  implicit none
  integer, private :: nx,ny,nf,nt,ireg
  real,    private :: theta,eps,dx,dy,dt,gscale
  logical, private :: verbose
  real,    private,allocatable :: alpha(:,:),beta(:,:),tmp(:,:)
  real,    private,allocatable :: l_f(:,:),m_f(:,:),s_f(:,:)
  real,    private,allocatable :: tmpx1(:),tmpx2(:),tmpy1(:),tmpy2(:)
  real,    private,allocatable :: A(:,:),B(:,:),C(:,:),F(:,:)
contains

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine meshing2D_init(ax,ay,af,eps_in,nt_in,dt_in,verbose_in,gscale_in,ireg_in)
    type(myaxis), intent(in) :: ax,ay,af
    real :: eps_in,dt_in,gscale_in
    integer :: nt_in,ireg_in
    logical :: verbose_in
    ireg=ireg_in
    gscale=gscale_in
    verbose=verbose_in
    nx=ax%n
    ny=ay%n
    nf=af%n
    nt=nt_in
    eps=eps_in
    dx=ax%d
    dy=ay%d
!    dx=(ax%d+ay%d)/2.
!    dy=dx
!    dx=1./real(nx-1)
!    dy=1./real(ny-1)
    dt=dt_in
write(0,*) dx,dy
    !! Note that dt < dx*dx/ (2 * D)
    !! where D is the diffusion coefficient
    !! Have to check to see if we are using a Crank-Nicholson method
    theta = dx*dx/dt; 
    allocate ( alpha(nx,ny),beta(nx,ny),tmp(nx,ny) )
    allocate ( l_f(nx,ny),m_f(nx,ny),s_f(nx,ny) )
    allocate (tmpx1(nx),tmpx2(nx),tmpy1(ny),tmpy2(ny) )
    allocate ( A(nx,ny),B(nx,ny),C(nx,ny),F(nx,ny) )
  end subroutine meshing2D_init

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine runmeshing2D(s,f)
    integer :: it
    real    :: f(:,:,:)
    complex :: s(:,:)
    real, allocatable,dimension(:,:) :: g11,g12,g22,g,m_u,m_v,p,q
    real, allocatable,dimension(:,:) :: a1,a4,a5
    real, allocatable,dimension(:,:) :: s_u,s_v,u,v,hu,hv,fu,fv
    real, allocatable :: b(:,:,:)
    logical :: to_break

    allocate( u(nx,ny),v(nx,ny),g11(nx,ny),g12(nx,ny),g22(nx,ny) )
    allocate( a1(nx,ny),a4(nx,ny),a5(nx,ny) )
    allocate( g(nx,ny),m_u(nx,ny),m_v(nx,ny),p(nx,ny),q(nx,ny),s_u(nx,ny) )
    allocate( b(nx,ny,nf),s_v(nx,ny),hu(nx,ny),hv(nx,ny),fu(nx,ny),fv(nx,ny) )

    !! Create coordinates 
    u = real ( s ) !! Direction 'x'
    v = aimag( s ) !! Direction 'y'

    !! Step 1
    !! Initalize Coordinates
    !! May redo for input coordinates!!
    !! Already done - call blend_coordinates(u,v)	

    !! Loop over solving parabolic equation
    !! ntau is the max. number of iterations

    f=0.
    !! Metric Calculations
    call calc_g11(u,v,f,g11)
    call calc_g12(u,v,f,g12)
    call calc_g22(u,v,f,g22)
    call calc_det(g,g11,g12,g22) 

    !! REGULARIZATION CHOICE
    if (ireg .eq. 1) then
       write(0,*) 'USING DETERMINANT'
       f=0.; f(2:nx-1,2:ny-1,1)= gscale * g(2:nx-1,2:ny-1)! / maxval(abs(g))
    else if (ireg .eq. 2) then
       write(0,*) 'USING G11'
       f=0.; f(2:nx-1,2:ny-1,1)= gscale * g11(2:nx-1,2:ny-1)! / maxval(abs(g))
    else if (ireg .eq. 3) then
       write(0,*) 'USING G12'
       f=0.; f(2:nx-1,2:ny-1,1)= gscale * g12(2:nx-1,2:ny-1)! / maxval(abs(g))
    else if (ireg .eq. 4) then
       write(0,*) 'USING G22'
       f=0.; f(2:nx-1,2:ny-1,1)= gscale * g22(2:nx-1,2:ny-1)! / maxval(abs(g))
    else
       write(0,*) 'USING DETERMINANT'
       f=0.; f(2:nx-1,2:ny-1,1)= gscale * g(2:nx-1,2:ny-1)! / maxval(abs(g))
    end if

    do it=1,nt
       write(0,*) 'Working on Step ',it,' of ',nt
       !! Step 2 - Calculate the essential ingredients of the
       !! diffusion equation solution process

       !! Metric Calculations
       call calc_g11(u,v,f,g11)
       if (verbose) write(0,*) 'G11 min and max',&
                minval(g11(2:nx-1,2:ny-1)),maxval(g11(2:nx-1,2:ny-1))
       call calc_g12(u,v,f,g12)
       if (verbose) write(0,*) 'G12 min and max',&
                minval(g12(2:nx-1,2:ny-1)),maxval(g12(2:nx-1,2:ny-1))
       call calc_g22(u,v,f,g22)
       if (verbose) write(0,*) 'G22 min and max',&
                minval(g22(2:nx-1,2:ny-1)),maxval(g22(2:nx-1,2:ny-1))

       !! Calculate determinant
       call calc_det(g,g11,g12,g22) 
       if (verbose) write(0,*) 'DET min/max',minval(g(2:nx-1,2:ny-1)),&
               maxval(g(2:nx-1,2:ny-1))
       !! REGULARIZATION CHOICE
       if (ireg .eq. 1) then
          f=0.; f(2:nx-1,2:ny-1,1)= gscale * g(2:nx-1,2:ny-1)! / maxval(abs(g))
       else if (ireg .eq. 2) then
          f=0.; f(2:nx-1,2:ny-1,1)= gscale * g11(2:nx-1,2:ny-1)! / maxval(abs(g))
       else if (ireg .eq. 3) then
          f=0.; f(2:nx-1,2:ny-1,1)= gscale * g12(2:nx-1,2:ny-1)! / maxval(abs(g))
       else if (ireg .eq. 4) then
          f=0.; f(2:nx-1,2:ny-1,1)= gscale * g22(2:nx-1,2:ny-1)! / maxval(abs(g))
       else
          f=0.; f(2:nx-1,2:ny-1,1)= gscale * g(2:nx-1,2:ny-1)! / maxval(abs(g))
       end if


       !   write(0,*) 'f min/max',minval(f),maxval(f)

       !! mixed derivatives
       call calc_mij(u,m_u)
       if (verbose) write(0,*) 'M_U :',minval(m_u),maxval(m_u)
       call calc_mij(v,m_v)
       if (verbose) write(0,*) 'M_V :',minval(m_v),maxval(m_v)

       !! Coefficients for monitor function
       call calc_bijk(g11,g12,g22,f,u,v,b)

       !! Monitor function gradients
       call calc_pij(f,u,b,p)
       call calc_qij(f,v,b,q)
       if (verbose) write(0,*) 'P :',minval(p),maxval(p)
       if (verbose) write(0,*) 'Q :',minval(q),maxval(q)

       !! Laplacian Derivatives
       call calc_sij(u,s_u)
       call calc_sij(v,s_v)
       if (verbose) write(0,*) 'S_V :',minval(s_v),maxval(s_v)
       if (verbose) write(0,*) 'S_U :',minval(s_u),maxval(s_u)

       !! Step 3 - Calculate u solution at 1/2 step
       !! in the x-direction
       call half_uij(g11,g12,g22,hu,u,m_u,s_u,p)
       if (verbose) write(0,*) 'u 1/2 max and min: ',minval(hu),maxval(hu)

       !! Step 4 - Calculate v solution at 1/2 step
       !! in the x-direction
       call half_vij(g11,g12,g22,hv,v,m_v,s_v,q)
       if (verbose) write(0,*) 'v 1/2 max and min: ',minval(hv),maxval(hv)

       !! Step 5 - Calculate u solution at full step
       !! in the y-direction
       call full_uij(g11,g12,g22,fu,hu,s_u)
       if (verbose) write(0,*) 'u max and min: ',minval(fu),maxval(fu)

       !! Step 6 -  Calculate v solution at full step 
       !! in the y-direction
       call full_vij(g11,g12,g22,fv,hv,s_v)
       if (verbose) write(0,*) 'v max and min: ',minval(fv),maxval(fv)

       !! Step 7 - Tolerance Check
       call tol_check(u,v,fu,fv,to_break)
       !! Step 8 - Reassign and iterate
       u=fu  
       v=fv
       if (to_break .eq. .true.) exit

    end do  !! End diffusion loop

    !! Output final coordinates
    s = cmplx(u,v)


    a1=0.;a4=0.;a5=0.;
    a1(2:nx-1,2:ny-1) = g12(2:nx-1,2:ny-1)**2/&
          (sqrt(g11(2:nx-1,2:ny-1)**2)*sqrt(g22(2:nx-1,2:ny-1)**2))
    a4(2:nx-1,2:ny-1) = sqrt(g(2:nx-1,2:ny-1)/g11(2:nx-1,2:ny-1))
    a5(2:nx-1,2:ny-1) = sqrt(g22(2:nx-1,2:ny-1)/g11(2:nx-1,2:ny-1) &
    - (g12(2:nx-1,2:ny-1)/g11(2:nx-1,2:ny-1))**2)

    a1(1,:)=a1(2,:); a1(nx,:)=a1(nx-1,:); a1(:,1)=a1(:,2); a1(:,ny)=a1(:,ny-1)
    a4(1,:)=a4(2,:); a4(nx,:)=a4(nx-1,:); a4(:,1)=a4(:,2); a4(:,ny)=a4(:,ny-1)
    a5(1,:)=a5(2,:); a5(nx,:)=a5(nx-1,:); a5(:,1)=a5(:,2); a5(:,ny)=a5(:,ny-1)

    call sep_write(g,"monitor.H")
    call sep_write(a1,"a1.H")
    call sep_write(a4,"a4.H")
    call sep_write(a5,"a5.H")

    deallocate ( g11,g12,g22,u,v,m_u,m_v,b,p,q,s_u,s_v,a1,a4,a5 )
  end subroutine runmeshing2D

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine calc_g11(u,v,f,g11)
    integer :: iif
    real,intent(in ) :: u(:,:),v(:,:),f(:,:,:)
    real,intent(out) :: g11(:,:)
    tmp=0.

    do iif=1,nf !! Do for multiple monitor functions
       tmp(2:nx-1,2:ny-1) = tmp(2:nx-1,2:ny-1) &
       +   ( f(3:nx,2:ny-1,iif) - f(1:nx-2,2:ny-1,iif) ) &
       *   ( f(3:nx,2:ny-1,iif) - f(1:nx-2,2:ny-1,iif) ) / (2.*dx)**2
    end do
!    tmp(1 ,: )=tmp(2   ,:   )
!    tmp(nx,: )=tmp(nx-1,:   )
!    tmp(: ,1 )=tmp(:   ,2   )
!    tmp(: ,ny)=tmp(:   ,ny-1)

    g11(2:nx-1,2:ny-1) = tmp + &
           ( ( u(3:nx,2:ny-1)-u(1:nx-2,2:ny-1) ) / (2.*dx) )**2 + &
           ( ( v(3:nx,2:ny-1)-v(1:nx-2,2:ny-1) ) / (2.*dx) )**2
!    g11(1,:) = g11(2,:); g11(nx,:) = g11(nx-1,:)
!    g11(:,1) = g11(:,2); g11(:,ny) = g11(:,ny-1)
 end subroutine calc_g11

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine calc_g12(u,v,f,g12)
    integer :: iif
    real,intent(in ) :: u(:,:),v(:,:),f(:,:,:)
    real,intent(out) :: g12(:,:)
    tmp=0.

    do iif=1,nf !! Do for multiple monitor functions
       tmp(2:nx-1,2:ny-1) = tmp(2:nx-1,2:ny-1) &
       +   ( f(3:nx  ,2:ny-1,iif) - f(1:nx-2,2:ny-1,iif) ) &
       *   ( f(2:nx-1,3:ny  ,iif) - f(2:nx-1,1:ny-2,iif) ) / (4.*dx*dy)
    end do
 !   tmp(1 ,: )=tmp(2   ,:   )
 !   tmp(nx,: )=tmp(nx-1,:   )
 !   tmp(: ,1 )=tmp(:   ,2   )
 !   tmp(: ,ny)=tmp(:   ,ny-1)

    g12(2:nx-1,2:ny-1) = tmp + &
         ( u(3:nx,2:ny-1)-u(1:nx-2,2:ny-1) ) * &
         ( u(2:nx-1,3:ny)-u(2:nx-1,1:ny-2) ) / (4.*dx*dy) + &
         ( v(3:nx,2:ny-1)-v(1:nx-2,2:ny-1) ) * &
         ( v(2:nx-1,3:ny)-v(2:nx-1,1:ny-2) ) / (4.*dx*dy)
!    g12(1,:) = g12(2,:); g12(nx,:) = g12(nx-1,:)
!    g12(:,1) = g12(:,2); g12(:,ny) = g12(:,ny-1)
  end subroutine calc_g12

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine calc_g22(u,v,f,g22)
    integer ::iif
    real,intent(in ) :: u(:,:),v(:,:),f(:,:,:)
    real,intent(out) :: g22(:,:)
    tmp=0.

    do iif=1,nf !! Do for multiple monitor functions
       tmp(2:nx-1,2:ny-1) = tmp(2:nx-1,2:ny-1) &
       +   ( f(2:nx-1,3:ny,iif) - f(2:nx-1,1:ny-2,iif) ) &
       *   ( f(2:nx-1,3:ny,iif) - f(2:nx-1,1:ny-2,iif) ) / (2.*dy)**2
    end do
!    tmp(1 ,: )=tmp(2   ,:   )
!    tmp(nx,: )=tmp(nx-1,:   )
!    tmp(: ,1 )=tmp(:   ,2   )
!    tmp(: ,ny)=tmp(:   ,ny-1)

    g22(2:nx-1,2:ny-1) = tmp + &
         ( ( u(2:nx-1,3:ny)-u(2:nx-1,1:ny-2) ) / (2.*dy) )**2 + &
         ( ( v(2:nx-1,3:ny)-v(2:nx-1,1:ny-2) ) / (2.*dy) )**2
!    g22(1,:) = g22(2,:); g22(nx,:) = g22(nx-1,:)
!    g22(:,1) = g22(:,2); g22(:,ny) = g22(:,ny-1)
  end subroutine calc_g22

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine calc_det(g,g11,g12,g22)
    real, intent(in ) :: g11(:,:),g12(:,:),g22(:,:)
    real, intent(out) :: g(:,:)
    g = g11 * g22 - g12 * g12
  end subroutine calc_det

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine calc_mij(s,m_s)
    real, intent(in ) ::   s(:,:)
    real, intent(out) :: m_s(:,:)
    m_s(2:nx-1,2:ny-1)= 0.25 * &
         ( s(3:nx  ,3:ny  ) &
          -s(1:nx-2,3:ny  ) &
          -s(3:nx  ,1:ny-2) &
          +s(1:nx-2,1:ny-2) )	

!    m_s(1,:) = m_s(2,:); m_s(nx,:) = m_s(nx-1,:)
!    m_s(:,1) = m_s(:,2); m_s(:,ny) = m_s(:,ny-1)
    m_s(1,:) = 0.; m_s(nx,:) = 0.
    m_s(:,1) = 0.; m_s(:,ny) = 0.

  end subroutine calc_mij

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine calc_lij(s,l_s)
    real, intent(in ) ::   s(:,:)
    real, intent(out) :: l_s(:,:)
    l_s(2:nx-1,2:ny-1)= (s(3:nx  ,2:ny-1) &
                     -2.*s(2:nx-1,2:ny-1) &
                        +s(1:nx-2,2:ny-1) )
 !   l_s(1,:) = l_s(2,:); l_s(nx,:) = l_s(nx-1,:)
 !   l_s(:,1) = l_s(:,2); l_s(:,ny) = l_s(:,ny-1)
    l_s(1,:) = 0.; l_s(nx,:) = 0.
    l_s(:,1) = 0.; l_s(:,ny) = 0.

  end subroutine calc_lij

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine calc_bijk(g11,g12,g22,f,u,v,b)
    integer :: iff
    real, intent(in ) :: g11(:,:),g12(:,:),g22(:,:)
    real, intent(in ) :: f(:,:,:),u(:,:),v(:,:)
    real, intent(out) :: b(:,:,:)
    l_f=0.; m_f=0.; s_f=0.

    do iff = 1,nf
       call calc_lij( f(:,:,iff) , l_f)
       call calc_mij( f(:,:,iff) , m_f)
       call calc_sij( f(:,:,iff) , s_f)
       b(:,:,iff) = g22 * l_f - &
                2.* g12 * m_f + &
                    g11 * s_f
    end do
!    b(1,:,:) = b(2,:,:); b(nx,:,:) = b(nx-1,:,:)
!    b(:,1,:) = b(:,2,:); b(:,ny,:) = b(:,ny-1,:)
    b(1,:,:) = 0.; b(nx,:,:) = 0.
    b(:,1,:) = 0.; b(:,ny,:) = 0.
  end subroutine calc_bijk

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine calc_pij(f,u,b,p)
    integer :: iif,iy,stat
    real, intent(in ) :: f(:,:,:),u(:,:),b(:,:,:)
    real, intent(out) :: p(:,:)
    tmpx1=0.; tmpx2=0.;

    do iy = 1,ny
    !   stat = igrad1_lop(.false.,.false., u(:,iy)   , tmpx2 )
       do iif = 1,nf
          stat = igrad1_lop(.false.,.false., f(:,iy,iif), tmpx1 )
          p(:,iy) = p(:,iy) + b(:,iy,iif)*tmpx1/dx !/(tmpx2+eps)
       end do
    end do
   p(1,:) = 0.; p(nx,:) = 0.
   p(:,1) = 0.; p(:,ny) = 0.

  end subroutine calc_pij

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine calc_qij(f,v,b,q)
    integer :: iif,ix,stat
    real, intent(in ) :: f(:,:,:),v(:,:),b(:,:,:)
    real, intent(out) :: q(:,:)
    tmpy1=0.; tmpy2=0.

    do ix = 1,nx
    !   stat = igrad1_lop(.false.,.false., v(ix,:)   , tmpy2 )
       do iif = 1,nf
          stat = igrad1_lop(.false.,.false., f(ix,:,iif), tmpy1 )
          q(ix,:) = q(ix,:) + b(ix,:,iif)*tmpy1/dy !/(tmpy2+eps)
       end do
    end do
   q(1,:) = 0.; q(nx,:) = 0.
   q(:,1) = 0.; q(:,ny) = 0.
  end subroutine calc_qij

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine calc_sij(s,s_s)
    real, intent(in ) :: s(:,:)
    real, intent(out) :: s_s(:,:)
    s_s(2:nx-1,2:ny-1)= (s(2:nx-1,3:ny )  &
                     -2.*s(2:nx-1,2:ny-1) &
                        +s(2:nx-1,1:ny-2) )	
  !  s_s(1,:) = s_s(2,:); s_s(nx,:) = s_s(nx-1,:)
  !  s_s(:,1) = s_s(:,2); s_s(:,ny) = s_s(:,ny-1)
   s_s(1,:) = 0.; s_s(nx,:) = 0.
   s_s(:,1) = 0.; s_s(:,ny) = 0.
  end subroutine calc_sij

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine half_uij(g11,g12,g22,hu,u,m_u,s_u,p) 	
    !! Calculate half-step in v coordinate w.r.t x
    integer :: ix 
    real, intent(in ) :: g11(:,:),g12(:,:),g22(:,:) 
    real, intent(in ) :: m_u(:,:),s_u(:,:),u(:,:),p(:,:)
    real, intent(out) :: hu(:,:) 
    alpha=0.;beta=0.;A=0.;B=0.;C=0.;F=0.; 

    !! Forward Calculate Coefficients
    alpha(1,:) = 0.; 
    beta (1,:) = u(1,:) 
    A = g22
    B = g22 
    C = 2.*g22 + theta
    F = theta*u -2.*g12*m_u + g11*s_u + p 
    if (verbose) write(0,*) 'u1 F min and max: ',minval(F),maxval(F) 

    do ix=1,nx-1 !! Note index shift 
       alpha(ix+1,:) = ( B(ix,:)                    ) / ( C(ix,:) - alpha(ix,:)*A(ix,:) )
       beta (ix+1,:) = ( A(ix,:)*beta(ix,:)+F(ix,:) ) / ( C(ix,:) - alpha(ix,:)*A(ix,:) )
    end do
    if (verbose) write(0,*) 'u1 beta min and max: ',minval(beta),maxval(beta)

    !! Back-filter coefficients
    hu (nx,:) = u(nx,:)
    do ix=nx-1,2,-1			
       hu (ix,:) = alpha(ix+1,:) * hu(ix+1,:) + beta(ix+1,:)
    end do
    hu(1 ,:)=u(1 ,:)
    hu(nx,:)=u(nx,:)

  end subroutine half_uij

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine half_vij(g11,g12,g22,hv,v,m_v,s_v,q)	
    !! Calculate half-step in v coordinate w.r.t x
    integer :: ix
    real, intent(in ) :: g11(:,:),g12(:,:),g22(:,:)
    real, intent(in ) :: m_v(:,:),s_v(:,:),v(:,:),q(:,:)
    real, intent(out) :: hv(:,:)
    alpha=0.;beta=0.;A=0.;B=0.;C=0.;F=0.;
    A = g22
    B = g22
    C = 2.*g22 + theta
    F = theta*v -2.*g12*m_v + g11*s_v + q
    if (verbose) write(0,*) 'v1 F min and max: ',minval(F),maxval(F)

    !! Forward Calculate Coefficients
    alpha(1,:) = 0.
    beta (1,:) = v(1,:)

    do ix=1,nx-1 !! Note index shift
       alpha(ix+1,:) = ( B(ix,:)                    ) / ( C(ix,:) - alpha(ix,:)*A(ix,:) )
       beta (ix+1,:) = ( A(ix,:)*beta(ix,:)+F(ix,:) ) / ( C(ix,:) - alpha(ix,:)*A(ix,:) )
    end do
    if (verbose) write(0,*) 'v1 beta min and max: ',minval(beta),maxval(beta)

    !! Back-filter coefficients
    hv (nx,:) = v(nx,:)
    do ix=nx-1,2,-1			
       hv (ix,:) = alpha(ix+1,:) * hv(ix+1,:) + beta(ix+1,:)
    end do
    hv(1 ,:) = v(1 ,:)
    hv(nx,:) = v(nx,:)
  end subroutine half_vij

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine full_uij(g11,g12,g22,fu,hu,s_u)
    !! Calculate half-step in u coordinate w.r.t y
    integer :: iy
    real, intent(in ) :: g11(:,:),g12(:,:),g22(:,:),s_u(:,:),hu(:,:)
    real, intent(out) :: fu(:,:)
    alpha=0.;beta=0.;A=0.;B=0.;C=0.;F=0.;
    A = g11
    B = g11
    C = 2.*g11 + theta
    F = theta*hu - g11*s_u
    if (verbose) write(0,*) 'u2 F min and max: ',minval(F),maxval(F)
    !! Forward Calculate Coefficients
    alpha(:,1) = 0.
    beta (:,1) = hu(:,1)

    do iy=1,ny-1
       alpha(:,iy+1)=( B(:,iy)                    ) / ( C(:,iy) - alpha(:,iy)*A(:,iy) )
       beta (:,iy+1)=( A(:,iy)*beta(:,iy)+F(:,iy) ) / ( C(:,iy) - alpha(:,iy)*A(:,iy) )
    end do
    if (verbose) write(0,*) 'u2 beta min and max: ',minval(beta),maxval(beta)

    !! Back filter coefficients
    fu (:,ny) = hu(:,ny)
    do iy=ny-1,2,-1			
       fu (:,iy) = alpha(:,iy+1) * fu(:,iy+1) + beta(:,iy+1)
    end do
    fu (:, 1) = hu(:, 1)
    fu (:,ny) = hu(:,ny)
  end subroutine full_uij

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine full_vij(g11,g12,g22,fv,hv,s_v)
    !! Calculate half-step in v coordinate w.r.t y
    integer :: iy
    real, intent(in ) :: g11(:,:),g12(:,:),g22(:,:),s_v(:,:),hv(:,:)
    real, intent(out) :: fv(:,:)
    alpha=0.;beta=0.;A=0.;B=0.;C=0.;F=0.;
    A = g11
    B = g11
    C = 2.*g11 + theta
    F = theta*hv - g11*s_v 
    if (verbose) write(0,*) 'v2 F min and max: ',minval(F),maxval(F)

    !! Forward Calculate Coefficients
    alpha(:,1) = 0.
    beta (:,1) = hv(:,1)
    do iy=1,ny-1
       alpha(:,iy+1)=( B(:,iy)                    ) / ( C(:,iy) - alpha(:,iy)*A(:,iy) )
       beta (:,iy+1)=( A(:,iy)*beta(:,iy)+F(:,iy) ) / ( C(:,iy) - alpha(:,iy)*A(:,iy) )
    end do
    if (verbose) write(0,*) 'v2 beta min and max: ',minval(beta),maxval(beta)

    !! Back filter coefficients
    fv (:,ny) = hv(:,ny)
    do iy=ny-1,2,-1			
       fv (:,iy) = alpha(:,iy+1) * fv(:,iy+1) + beta(:,iy+1)
    end do
    fv (:, 1) = hv(:, 1)
    fv (:,ny) = hv(:,ny)
  end subroutine full_vij

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine tol_check(u,v,fu,fv,to_break)
    real   ,intent(in ):: u(:,:),v(:,:),fu(:,:),fv(:,:)
    logical,intent(out):: to_break

!!    if (maxval ( sqrt( (fu-u)**2 - (fv-v)**2 ))  .lt. eps ) then
    if (maxval ( 1/dt * sqrt( (fu-u)**2 - (fv-v)**2 ))  .lt. eps ) then
       to_break = .true.
    else
       to_break = .false.
    end if
!write(0,*) 'Change is :',maxval(sqrt((fu-u)**2-(fv-v)**2)),' relative to desired ',eps
!if (verbose) write(0,*) " "
!if (verbose) write(0,*) " "
!if (verbose) write(0,*) " "

  end subroutine tol_check

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

end module meshing2D

